from typing import BinaryIO
import base64

from kloppy.config import get_config
from kloppy.exceptions import AdapterError
from .adapter import Adapter


class HTTPAdapter(Adapter):
    def supports(self, url: str) -> bool:
        return url.startswith("http://") or url.startswith("https://")

    def read_to_stream(self, url: str, output: BinaryIO):
        basic_authentication = get_config("adapters.http.basic_authentication")

        try:
            from js import XMLHttpRequest

            _RUNS_IN_BROWSER = True
        except ImportError:
            try:
                import requests
            except ImportError:
                raise AdapterError(
                    "Seems like you don't have requests installed. Please"
                    " install it using: pip install requests"
                )

            _RUNS_IN_BROWSER = False

        if _RUNS_IN_BROWSER:
            request = XMLHttpRequest.new()
            if basic_authentication:
                authentication = base64.b64encode(
                    basic_authentication.join(":")
                )
                request.setRequestHeader(
                    "Authorization",
                    f"Basic {authentication}",
                )

            request.open("GET", url, False)
            request.send(None)
            output.write(request.responseText)
        else:
            auth = None
            if basic_authentication:
                auth = requests.auth.HTTPBasicAuth(*basic_authentication)

            with requests.get(url, stream=True, auth=auth) as r:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=8192):
                    output.write(chunk)



import json
import logging
from typing import Dict, List, Tuple, Union, IO, NamedTuple

from kloppy.domain import (
    AttackingDirection,
    BallOutEvent,
    BallState,
    BodyPart,
    BodyPartQualifier,
    CardEvent,
    CardType,
    DatasetFlag,
    Event,
    EventDataset,
    FoulCommittedEvent,
    GenericEvent,
    Ground,
    Metadata,
    Orientation,
    PassEvent,
    PassResult,
    Period,
    Player,
    Point,
    Provider,
    Qualifier,
    RecoveryEvent,
    Score,
    SetPieceQualifier,
    SetPieceType,
    ShotEvent,
    ShotResult,
    SubstitutionEvent,
    Team,
)
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.deserializer import EventDataDeserializer
from kloppy.utils import Readable, performance_logging


logger = logging.getLogger(__name__)


DF_EVENT_CLASS_STATUS = "status"
DF_EVENT_CLASS_NUTMEGS = "nutmegs"
DF_EVENT_CLASS_STEALINGS = "stealings"
DF_EVENT_CLASS_FOULS = "fouls"
DF_EVENT_CLASS_THROW_IN = "throwIn"
DF_EVENT_CLASS_INCORRECT_PASSES = "incorrectPasses"
DF_EVENT_CLASS_CLEARANCES = "clearances"
DF_EVENT_CLASS_CORRECT_PASSES = "correctPasses"
DF_EVENT_CLASS_SOMBRERO_FLICK = "sombreroFlick"
DF_EVENT_CLASS_GOALKICK = "goalkick"
DF_EVENT_CLASS_GOALS = "goals"
DF_EVENT_CLASS_SHOTS = "shots"
DF_EVENT_CLASS_OFFSIDES = "offsides"
DF_EVENT_CLASS_RED_CARDS = "redCards"
DF_EVENT_CLASS_YELLOW_CARDS = "yellowCards"
DF_EVENT_CLASS_SUBSTITUTIONS = "substitutions"
DF_EVENT_CLASS_PENALTY_SHOOTOUT = "penaltyShootout"
DF_EVENT_CLASS_CORNERKICKS = "cornerKicks"

DF_EVENT_CLASS_PASSES = {
    DF_EVENT_CLASS_THROW_IN,
    DF_EVENT_CLASS_INCORRECT_PASSES,
    DF_EVENT_CLASS_CORRECT_PASSES,
    DF_EVENT_CLASS_CORNERKICKS,
}
DF_EVENT_CLASS_CARDS = {DF_EVENT_CLASS_YELLOW_CARDS, DF_EVENT_CLASS_RED_CARDS}

# status event types
DF_EVENT_TYPE_STATUS_MATCH_START = 1
DF_EVENT_TYPE_STATUS_MATCH_END = 2
DF_EVENT_TYPE_STATUS_FIRST_HALF_END = 17
DF_EVENT_TYPE_STATUS_SECOND_HALF_START = 18
DF_EVENT_TYPE_STATUS_SECOND_HALF_END = 49
DF_EVENT_TYPE_STATUS_FIRST_EXTRA_START = 50
DF_EVENT_TYPE_STATUS_FIRST_EXTRA_END = 51
DF_EVENT_TYPE_STATUS_SECOND_EXTRA_START = 52
DF_EVENT_TYPE_STATUS_SECOND_EXTRA_END = 53
DF_EVENT_TYPE_STATUS_PENALTY_SHOOTOUT_START = 54

# card event types
DF_EVENT_TYPE_YELLOW_CARD = 3
DF_EVENT_TYPE_RED_CARD_DOUBLE_YELLOW = 4
DF_EVENT_TYPE_RED_CARD = 5

# shots
DF_EVENT_TYPE_SHOT_OPEN_PLAY_GOAL = 9
DF_EVENT_TYPE_SHOT_OWN_GOAL = 10
DF_EVENT_TYPE_SHOT_HEAD_GOAL = 11
DF_EVENT_TYPE_SHOT_FREE_KICK_GOAL = 12
DF_EVENT_TYPE_SHOT_PENALTY_GOAL = 13
DF_EVENT_TYPE_SHOT_OFF_TARGET = 33
DF_EVENT_TYPE_SHOT_POST = 34
DF_EVENT_TYPE_SHOT_SAVED = 35
DF_EVENT_TYPE_SHOT_PENALTY_SAVED = 43
DF_EVENT_TYPE_SHOT_PENALTY_OFF_TARGET = 44
DF_EVENT_TYPE_SHOT_PENALTY_POST = 45

GOAL_EVENTS = {
    DF_EVENT_TYPE_SHOT_OPEN_PLAY_GOAL,
    DF_EVENT_TYPE_SHOT_OWN_GOAL,
    DF_EVENT_TYPE_SHOT_HEAD_GOAL,
    DF_EVENT_TYPE_SHOT_FREE_KICK_GOAL,
    DF_EVENT_TYPE_SHOT_PENALTY_GOAL,
}

PENALTY_SHOT_EVENTS = {
    DF_EVENT_TYPE_SHOT_PENALTY_GOAL,
    DF_EVENT_TYPE_SHOT_PENALTY_SAVED,
    DF_EVENT_TYPE_SHOT_PENALTY_OFF_TARGET,
    DF_EVENT_TYPE_SHOT_PENALTY_POST,
}

# ball out events
DF_EVENT_TYPE_CORNER_KICK = 28
DF_EVENT_TYPE_THROW_IN = 46
DF_EVENT_TYPE_GOAL_KICK = 48

BALL_OUT_EVENTS = {
    DF_EVENT_TYPE_THROW_IN,
    DF_EVENT_TYPE_CORNER_KICK,
    DF_EVENT_TYPE_GOAL_KICK,
}

# stealings
DF_EVENT_TYPE_CLEARANCE = 29
DF_EVENT_TYPE_CLEARANCE_TO_CORNER = 30
DF_EVENT_TYPE_STEALING = 182

# fouls
DF_EVENT_TYPE_FOUL_PENALTY = 31
DF_EVENT_TYPE_HAND_PENALTY = 32
DF_EVENT_TYPE_FOUL = 36
DF_EVENT_TYPE_HAND = 37
DF_EVENT_TYPE_OFFSIDE = 47

FOUL_FREE_KICK_EVENTS = {
    DF_EVENT_TYPE_FOUL,
    DF_EVENT_TYPE_HAND,
}

# passes
DF_EVENT_TYPE_PASS_CORRECT = 180
DF_EVENT_TYPE_PASS_INCORRECT = 181

# misc
DF_EVENT_TYPE_NUTMEG = 507
DF_EVENT_TYPE_SOMBRERO_FLICK = 509
DF_EVENT_TYPE_PENALTY_SHOOTOUT_GOAL = 55
DF_EVENT_TYPE_PENALTY_SHOOTOUT_SAVED = 56
DF_EVENT_TYPE_PENALTY_SHOOTOUT_OFF_TARGET = 57
DF_EVENT_TYPE_PENALTY_SHOOTOUT_POST = 183


def parse_str_ts(raw_event: Dict) -> float:
    return raw_event["t"]["m"] * 60 + (raw_event["t"]["s"] or 0)


def _parse_coordinates(coordinates: Dict[str, float]) -> Point:
    # location is cell based
    # +-------+-------+
    # | -1,-1 |  1,-1 |
    # +-------+-------+
    # | -1,1  |  1,1  |
    # +-------+-------+
    return Point(x=coordinates["x"], y=coordinates["y"])


def _get_team_and_player(
    raw_event: Dict, home_team: Team, away_team: Team
) -> Tuple[Team, Player]:
    team = None
    player = None

    team_id = raw_event.get("team")
    if team_id is not None:
        team = home_team if str(team_id) == home_team.team_id else away_team

    if "plyrId" in raw_event:
        if team is not None:
            player = team.get_player_by_id(raw_event["plyrId"])
        else:
            # NOTE: sometime events are missing team
            player = home_team.get_player_by_id(raw_event["plyrId"])
            if player is not None:
                team = home_team
            else:
                player = away_team.get_player_by_id(raw_event["plyrId"])
                team = away_team
    return team, player


def _get_event_qualifiers(
    raw_event: Dict, previous_event: Dict = None
) -> List[Qualifier]:
    qualifiers = []

    if raw_event["type"] == DF_EVENT_TYPE_THROW_IN:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.THROW_IN))

    if raw_event["type"] == DF_EVENT_TYPE_CORNER_KICK:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.CORNER_KICK))

    if raw_event["type"] in PENALTY_SHOT_EVENTS:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.PENALTY))

    if raw_event["type"] == DF_EVENT_TYPE_SHOT_HEAD_GOAL:
        qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))

    if (
        previous_event is not None
        and previous_event["type"] == DF_EVENT_TYPE_GOAL_KICK
        and previous_event["plyrId"] == raw_event["plyrId"]
    ):
        qualifiers.append(SetPieceQualifier(value=SetPieceType.GOAL_KICK))

    if raw_event["type"] == DF_EVENT_TYPE_SHOT_FREE_KICK_GOAL or (
        previous_event is not None
        and previous_event["type"] in FOUL_FREE_KICK_EVENTS
    ):
        qualifiers.append(SetPieceQualifier(value=SetPieceType.FREE_KICK))

    if previous_event is None or previous_event["type"] in GOAL_EVENTS:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.KICK_OFF))

    return qualifiers


def _parse_pass(
    raw_event: Dict,
    team: Team,
    previous_event: Dict = None,
    next_event: Dict = None,
) -> Dict:
    if next_event is not None and next_event["type"] == DF_EVENT_TYPE_OFFSIDE:
        result = PassResult.OFFSIDE
    elif raw_event["type"] == DF_EVENT_TYPE_PASS_CORRECT:
        result = PassResult.COMPLETE
    elif next_event is not None and next_event["type"] in BALL_OUT_EVENTS:
        result = PassResult.OUT
    else:
        result = PassResult.INCOMPLETE

    receiver_coordinates = None
    # NOTE: in really rare cases coordinates may be missing
    coordinates = raw_event.get("coord", {}).get("2")
    if coordinates is not None:
        receiver_coordinates = _parse_coordinates(coordinates)

    # find receiver player, if possible
    receiver_player = None
    if "recvId" in raw_event:
        receiver_player = team.get_player_by_id(raw_event["recvId"])

    elif raw_event["type"] in (
        DF_EVENT_TYPE_THROW_IN,
        DF_EVENT_TYPE_CORNER_KICK,
    ):
        # there is no receiver data in this case, use next_event to deduce receiver
        if next_event is not None:
            if (
                next_event["team"] != raw_event["team"]
                and next_event["type"] == DF_EVENT_TYPE_FOUL
                and "recvId" in next_event
            ):
                # foul from the opposite team
                receiver_player = team.get_player_by_id(next_event["recvId"])
                result = PassResult.COMPLETE
            elif (
                next_event["team"] == raw_event["team"]
                and "plyrId" in next_event
            ):
                # or same team event with plyrId
                receiver_player = team.get_player_by_id(next_event["plyrId"])
                result = PassResult.COMPLETE

    qualifiers = _get_event_qualifiers(raw_event, previous_event)

    return dict(
        result=result,
        receiver_coordinates=receiver_coordinates,
        receiver_player=receiver_player,
        receive_timestamp=parse_str_ts(raw_event),
        qualifiers=qualifiers,
    )


def _parse_shot(raw_event: Dict, previous_event: Dict = None) -> Dict:
    outcome_id = raw_event["type"]
    if outcome_id in GOAL_EVENTS:
        result = ShotResult.GOAL
    elif outcome_id in (
        DF_EVENT_TYPE_SHOT_OFF_TARGET,
        DF_EVENT_TYPE_SHOT_PENALTY_OFF_TARGET,
    ):
        result = ShotResult.OFF_TARGET
    elif outcome_id in (
        DF_EVENT_TYPE_SHOT_POST,
        DF_EVENT_TYPE_SHOT_PENALTY_POST,
    ):
        result = ShotResult.POST
    elif outcome_id in (
        DF_EVENT_TYPE_SHOT_SAVED,
        DF_EVENT_TYPE_SHOT_PENALTY_SAVED,
    ):
        result = ShotResult.SAVED
    else:
        raise DeserializationError(f"Unknown shot outcome: {outcome_id}")

    qualifiers = _get_event_qualifiers(raw_event, previous_event)

    return dict(
        result=result,
        qualifiers=qualifiers,
    )


def _parse_card(raw_event: Dict) -> Dict:
    card_id = raw_event["type"]
    if card_id == DF_EVENT_TYPE_RED_CARD:
        card_type = CardType.RED
    elif card_id == DF_EVENT_TYPE_RED_CARD_DOUBLE_YELLOW:
        card_type = CardType.SECOND_YELLOW
    elif card_id == DF_EVENT_TYPE_YELLOW_CARD:
        card_type = CardType.FIRST_YELLOW
    else:
        raise DeserializationError(f"Unknown card id {card_id}")

    return dict(card_type=card_type)


def _parse_substitution(raw_event: Dict, team: Team) -> Dict:
    player = team.get_player_by_id(raw_event["offId"])
    replacement_player = team.get_player_by_id(raw_event["inId"])

    return dict(player=player, replacement_player=replacement_player)


def _include_event(event: Event, wanted_event_types: List) -> bool:
    return not wanted_event_types or event.event_type in wanted_event_types


class DatafactoryInputs(NamedTuple):
    event_data: IO[bytes]


class DatafactoryDeserializer(EventDataDeserializer[DatafactoryInputs]):
    @property
    def provider(self) -> Provider:
        return Provider.DATAFACTORY

    def deserialize(self, inputs: DatafactoryInputs) -> EventDataset:

        transformer = self.get_transformer(length=2, width=2)

        with performance_logging("load data", logger=logger):
            data = json.load(inputs.event_data)
            match = data["match"]
            score_data = data["scoreStatus"]
            incidences = data["incidences"]
            players_data = data["players"]
            teams_data = data["teams"]

        with performance_logging("parse data", logger=logger):
            teams = []
            scores = []
            team_ids = (
                (Ground.HOME, str(match["homeTeamId"])),
                (Ground.AWAY, str(match["awayTeamId"])),
            )
            for ground, team_id in team_ids:
                team = Team(
                    team_id=team_id,
                    name=teams_data[team_id]["name"],
                    ground=ground,
                )
                team.players = [
                    Player(
                        player_id=player_id,
                        team=team,
                        first_name=player["name"]["first"],
                        last_name=player["name"]["last"],
                        name=player["name"]["shortName"]
                        or player["name"]["nick"],
                        jersey_no=player["squadNo"],
                        starting=not player["substitute"],
                    )
                    for player_id, player in players_data.items()
                    if str(player["teamId"]) == team_id
                ]
                teams.append(team)
                scores.append(score_data.get(team_id, {}).get("score"))
            score = Score(home=scores[0], away=scores[1])

            # setup periods
            status = incidences.pop(DF_EVENT_CLASS_STATUS)
            # start timestamps are fixed
            start_ts = {1: 0, 2: 45 * 60, 3: 90 * 60, 4: 105 * 60, 5: 120 * 60}
            # check for end status updates to setup periods
            end_event_types = {
                DF_EVENT_TYPE_STATUS_MATCH_END,
                DF_EVENT_TYPE_STATUS_FIRST_HALF_END,
                DF_EVENT_TYPE_STATUS_SECOND_HALF_END,
                DF_EVENT_TYPE_STATUS_FIRST_EXTRA_END,
                DF_EVENT_TYPE_STATUS_SECOND_EXTRA_END,
            }
            periods = {}
            for status_update in status.values():
                if status_update["type"] not in end_event_types:
                    continue
                half = status_update["t"]["half"]
                end_ts = parse_str_ts(status_update)
                periods[half] = Period(
                    id=half,
                    start_timestamp=start_ts[half],
                    end_timestamp=end_ts,
                    attacking_direction=AttackingDirection.HOME_AWAY
                    if half % 2 == 1
                    else AttackingDirection.AWAY_HOME,
                )

            # exclude goals, already listed as shots too
            incidences.pop(DF_EVENT_CLASS_GOALS)
            raw_events = [
                (k, e_id, e)
                for k in incidences
                for e_id, e in incidences[k].items()
            ]
            # sort events by timestamp, event_id
            raw_events.sort(
                key=lambda e: (
                    e[2]["t"]["half"],
                    e[2]["t"]["m"],
                    e[2]["t"]["s"] or 0,
                    e[1],
                )
            )

            home_team, away_team = teams
            events = []
            previous_event = next_event = None
            for i, (e_class, e_id, raw_event) in enumerate(raw_events):
                period = periods.get(raw_event["t"]["half"])
                if period is None:
                    # skip invalid event
                    continue

                timestamp = parse_str_ts(raw_event)
                if (
                    previous_event is not None
                    and previous_event["t"]["half"] != raw_event["t"]["half"]
                ):
                    previous_event = None
                next_event = (
                    raw_events[i + 1][2] if i + 1 < len(raw_events) else None
                )

                team, player = _get_team_and_player(
                    raw_event, home_team, away_team
                )

                event_base_kwargs = dict(
                    # from DataRecord
                    period=period,
                    timestamp=timestamp,
                    ball_owning_team=team,
                    ball_state=BallState.ALIVE,
                    # from Event
                    event_id=e_id,
                    team=team,
                    player=player,
                    coordinates=(
                        _parse_coordinates(raw_event["coord"]["1"])
                        if "coord" in raw_event
                        else None
                    ),
                    raw_event=raw_event,
                    result=None,
                    qualifiers=None,
                )

                if e_class in DF_EVENT_CLASS_PASSES:
                    pass_event_kwargs = _parse_pass(
                        raw_event=raw_event,
                        team=team,
                        previous_event=previous_event,
                        next_event=next_event,
                    )
                    event_base_kwargs.update(pass_event_kwargs)
                    event = PassEvent.create(**event_base_kwargs)

                elif e_class == DF_EVENT_CLASS_SHOTS:
                    shot_event_kwargs = _parse_shot(
                        raw_event=raw_event,
                        previous_event=previous_event,
                    )
                    event_base_kwargs.update(shot_event_kwargs)
                    event = ShotEvent.create(**event_base_kwargs)

                elif e_class == DF_EVENT_CLASS_STEALINGS:
                    event = RecoveryEvent.create(**event_base_kwargs)

                elif e_class == DF_EVENT_CLASS_FOULS:
                    # NOTE: could use qualifiers? (hand, foul, penalty?)
                    # switch possession team
                    event_base_kwargs["ball_owning_team"] = (
                        home_team if team == away_team else away_team
                    )
                    event = FoulCommittedEvent.create(**event_base_kwargs)

                elif e_class in DF_EVENT_CLASS_CARDS:
                    card_kwargs = _parse_card(
                        raw_event=raw_event,
                    )
                    event_base_kwargs.update(card_kwargs)
                    event = CardEvent.create(**event_base_kwargs)

                elif e_class == DF_EVENT_CLASS_SUBSTITUTIONS:
                    substitution_event_kwargs = _parse_substitution(
                        raw_event=raw_event, team=team
                    )
                    event_base_kwargs.update(substitution_event_kwargs)
                    event = SubstitutionEvent.create(**event_base_kwargs)

                else:
                    # otherwise, a generic event
                    event = GenericEvent.create(
                        event_name=e_class,
                        **event_base_kwargs,
                    )

                # check if the event implies ball was out of the field and add a synthetic out event
                if raw_event["type"] in BALL_OUT_EVENTS:
                    ball_out_event = BallOutEvent.create(
                        # from DataRecord
                        period=period,
                        timestamp=timestamp,
                        ball_owning_team=team,
                        ball_state=BallState.DEAD,
                        # from Event
                        event_id=e_id,
                        team=team,
                        player=player,
                        coordinates=event.coordinates,
                        raw_event=raw_event,
                        result=None,
                        qualifiers=None,
                    )
                    if self.should_include_event(event):
                        events.append(
                            transformer.transform_event(ball_out_event)
                        )

                if self.should_include_event(event):
                    events.append(transformer.transform_event(event))

                # only consider as a previous_event a ball-in-play event
                if e_class not in (
                    DF_EVENT_CLASS_YELLOW_CARDS,
                    DF_EVENT_CLASS_RED_CARDS,
                    DF_EVENT_CLASS_SUBSTITUTIONS,
                    DF_EVENT_CLASS_PENALTY_SHOOTOUT,
                ):
                    previous_event = raw_event

        metadata = Metadata(
            teams=teams,
            periods=sorted(periods.values(), key=lambda p: p.id),
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            frame_rate=None,
            orientation=Orientation.HOME_TEAM,
            flags=DatasetFlag.BALL_OWNING_TEAM,
            score=score,
            provider=Provider.DATAFACTORY,
            coordinate_system=transformer.get_to_coordinate_system(),
        )

        return EventDataset(
            metadata=metadata,
            records=events,
        )



import os
from contextlib import contextmanager
from copy import copy
from typing import Any, Optional

try:
    from typing import TypedDict
except ImportError:
    from mypy_extensions import TypedDict


cache_dir = os.environ.get("KLOPPY_CACHE_DIR")
if not cache_dir:
    cache_dir = os.path.expanduser("~/kloppy_cache")

Config = TypedDict(
    "Config",
    {
        "cache": Optional[str],
        "coordinate_system": Optional[str],
        "adapters.http.basic_authentication": Optional[str],
        "adapters.s3.s3fs": Optional[Any],
    },
)


class PartialConfig(Config, total=False):
    pass


_default_config: Config = {
    "cache": cache_dir,
    "coordinate_system": "kloppy",
    "adapters.http.basic_authentication": None,
    "adapters.s3.s3fs": None,
}

config = copy(_default_config)


def reset_config():
    for key, value in _default_config.items():
        set_config(key, value)


def set_config(key: str, value: Optional[str]):
    if key in config:
        config[key] = value
    else:
        raise KeyError(f"Non existing config '{key}'")


def get_config(key: Optional[str] = None):
    if key is None:
        return config
    elif key in config:
        return config[key]
    else:
        raise KeyError(f"Non existing config '{key}'")


@contextmanager
def config_context(*args):
    """Set some config items for within a certain context. Code borrowed partly from
    pandas."""
    if len(args) % 2 != 0 or len(args) < 2:
        raise ValueError(
            "Need to invoke as config_context(key, value, [(key, value), ...])."
        )

    configs = list(zip(args[::2], args[1::2]))

    undo = {key: get_config(key) for key, _ in configs}
    try:
        for key, value in configs:
            set_config(key, value)

        yield

    finally:
        for key, value in undo.items():
            set_config(key, value)



from abc import ABC, abstractmethod
import math
from typing import Dict, Any, Set, Type, Union, List, Optional

from kloppy.domain import (
    Event,
    BodyPartQualifier,
    DatasetTransformer,
    Orientation,
)
from kloppy.domain.models.event import (
    EnumQualifier,
    CarryEvent,
    PassEvent,
    EventType,
    ShotEvent,
    CardEvent,
)
from kloppy.exceptions import (
    UnknownEncoderError,
    OrientationError,
    KloppyParameterError,
)
from kloppy.utils import camelcase_to_snakecase


class OneHotEncoder:
    def __init__(self, name: str, options: List[str]):
        self.options = options
        self.name = name

    def encode(self, value: Union[Set, str]) -> Dict[str, bool]:
        if isinstance(value, str):
            value = [value]

        return {
            f"is_{self.name}_{option_.lower()}": option_ in value
            for option_ in self.options
        }


class EventAttributeTransformer(ABC):
    @abstractmethod
    def __call__(self, event: Event) -> Dict[str, Any]:
        pass


class AngleToGoalTransformer(EventAttributeTransformer):
    def __call__(self, event: Event) -> Dict[str, Any]:
        metadata = event.dataset.metadata
        if metadata.orientation != Orientation.ACTION_EXECUTING_TEAM:
            raise OrientationError(
                "Can only calculate Angle when dataset orientation is ACTION_EXECUTING_TEAM"
            )

        if not event.coordinates:
            return {"angle_to_goal": None}

        if metadata.pitch_dimensions.width:
            # Calculate in metric system
            event_x = event.coordinates.x * metadata.pitch_dimensions.length
            event_y = event.coordinates.y * metadata.pitch_dimensions.width
            goal_x = metadata.pitch_dimensions.length
            goal_y = metadata.pitch_dimensions.width / 2
        else:
            event_x = event.coordinates.x
            event_y = event.coordinates.y
            goal_x = metadata.pitch_dimensions.x_dim.max
            goal_y = (
                metadata.pitch_dimensions.y_dim.max
                + metadata.pitch_dimensions.y_dim.min
            ) / 2

        return {
            "angle_to_goal": math.atan2(goal_x - event_x, goal_y - event_y)
            / math.pi
            * 180
        }


class DistanceToGoalTransformer(EventAttributeTransformer):
    def __call__(self, event: Event) -> Dict[str, Any]:
        metadata = event.dataset.metadata
        if metadata.orientation != Orientation.ACTION_EXECUTING_TEAM:
            raise OrientationError(
                "Can only calculate Angle when dataset orientation is ACTION_EXECUTING_TEAM"
            )

        if not event.coordinates:
            return {"distance_to_goal": None}

        event_x = event.coordinates.x
        event_y = event.coordinates.y
        goal_x = metadata.pitch_dimensions.x_dim.max
        goal_y = (
            metadata.pitch_dimensions.y_dim.max
            + metadata.pitch_dimensions.y_dim.min
        ) / 2

        return {
            "distance_to_goal": math.sqrt(
                (goal_x - event_x) ** 2 + (goal_y - event_y) ** 2
            )
        }


def create_transformer_from_qualifier(
    qualifier_type: Type[EnumQualifier],
) -> Type[EventAttributeTransformer]:
    enum_ = qualifier_type.__annotations__["value"]
    name = camelcase_to_snakecase(enum_.__name__)
    options = [e.value for e in enum_]

    class _Transformer(EventAttributeTransformer):
        def __init__(self, encoding: str = "one-hot"):
            if encoding == "one-hot":
                self.encoder = OneHotEncoder(name, options)
            else:
                raise UnknownEncoderError(f"Don't know {encoding} encoding")

        def __call__(self, event: Event) -> Dict[str, Any]:
            values = {
                qualifier.value.value
                for qualifier in event.qualifiers
                if isinstance(qualifier, qualifier_type)
            }
            return self.encoder.encode(values)

    return _Transformer


class DefaultTransformer(EventAttributeTransformer):
    def __init__(
        self,
        *include: str,
        exclude: Optional[List[str]] = None,
    ):
        if include and exclude:
            raise KloppyParameterError(
                f"Cannot specify both include as exclude"
            )

        self.exclude = exclude or []
        self.include = include or []

    def __call__(self, event: Event) -> Dict[str, Any]:
        row = dict(
            event_id=event.event_id,
            event_type=(
                event.event_type.value
                if event.event_type != EventType.GENERIC
                else f"GENERIC:{event.event_name}"
            ),
            result=event.result.value if event.result else None,
            success=event.result.is_success if event.result else None,
            period_id=event.period.id,
            timestamp=event.timestamp,
            end_timestamp=None,
            ball_state=event.ball_state.value if event.ball_state else None,
            ball_owning_team=event.ball_owning_team.team_id
            if event.ball_owning_team
            else None,
            team_id=event.team.team_id if event.team else None,
            player_id=event.player.player_id if event.player else None,
            coordinates_x=event.coordinates.x if event.coordinates else None,
            coordinates_y=event.coordinates.y if event.coordinates else None,
        )
        if isinstance(event, PassEvent):
            row.update(
                {
                    "end_timestamp": event.receive_timestamp,
                    "end_coordinates_x": event.receiver_coordinates.x
                    if event.receiver_coordinates
                    else None,
                    "end_coordinates_y": event.receiver_coordinates.y
                    if event.receiver_coordinates
                    else None,
                    "receiver_player_id": event.receiver_player.player_id
                    if event.receiver_player
                    else None,
                }
            )
        elif isinstance(event, CarryEvent):
            row.update(
                {
                    "end_timestamp": event.end_timestamp,
                    "end_coordinates_x": event.end_coordinates.x
                    if event.end_coordinates
                    else None,
                    "end_coordinates_y": event.end_coordinates.y
                    if event.end_coordinates
                    else None,
                }
            )
        elif isinstance(event, ShotEvent):
            row.update(
                {
                    "end_coordinates_x": event.result_coordinates.x
                    if event.result_coordinates
                    else None,
                    "end_coordinates_y": event.result_coordinates.y
                    if event.result_coordinates
                    else None,
                }
            )
        elif isinstance(event, CardEvent):
            row.update(
                {
                    "card_type": event.card_type.value
                    if event.card_type
                    else None
                }
            )

        if event.qualifiers:
            for qualifier in event.qualifiers:
                row.update(qualifier.to_dict())

        if self.include:
            return {k: row[k] for k in self.include}
        elif self.exclude:
            return {k: v for k, v in row.items() if k not in self.exclude}
        else:
            return row


BodyPartTransformer = create_transformer_from_qualifier(BodyPartQualifier)



import logging
from typing import NamedTuple, IO
from dataclasses import replace

from kloppy.domain import (
    TrackingDataset,
    Frame,
    Point,
    Point3D,
    Provider,
    PlayerData,
    DatasetTransformer,
)
from kloppy.utils import performance_logging

from .metadata import load_metadata, EPTSMetadata
from .reader import read_raw_data
from ..deserializer import TrackingDataDeserializer

logger = logging.getLogger(__name__)


class MetricaEPTSTrackingDataInputs(NamedTuple):
    meta_data: IO[bytes]
    raw_data: IO[bytes]


class MetricaEPTSTrackingDataDeserializer(
    TrackingDataDeserializer[MetricaEPTSTrackingDataInputs]
):
    @property
    def provider(self) -> Provider:
        return Provider.METRICA

    @staticmethod
    def _frame_from_row(
        row: dict, metadata: EPTSMetadata, transformer: DatasetTransformer
    ) -> Frame:
        timestamp = row["timestamp"]
        if metadata.periods and row["period_id"]:
            # might want to search for it instead
            period = metadata.periods[row["period_id"] - 1]
        else:
            period = None

        other_sensors = []
        for sensor in metadata.sensors:
            if sensor.sensor_id not in ["position", "distance", "speed"]:
                other_sensors.append(sensor)

        players_data = {}
        for team in metadata.teams:
            for player in team.players:

                other_data = {}
                for sensor in other_sensors:
                    other_data.update(
                        {
                            sensor.sensor_id: row[
                                f"player_{player.player_id}_{sensor.channels[0].channel_id}"
                            ]
                        }
                    )

                players_data[player] = PlayerData(
                    coordinates=Point(
                        x=row[f"player_{player.player_id}_x"],
                        y=row[f"player_{player.player_id}_y"],
                    )
                    if f"player_{player.player_id}_x" in row
                    else None,
                    speed=row[f"player_{player.player_id}_s"]
                    if f"player_{player.player_id}_s" in row
                    else None,
                    distance=row[f"player_{player.player_id}_d"]
                    if f"player_{player.player_id}_d" in row
                    else None,
                    other_data=other_data,
                )

        frame = Frame(
            frame_id=row["frame_id"],
            timestamp=timestamp,
            ball_owning_team=None,
            ball_state=None,
            period=period,
            players_data=players_data,
            other_data={},
            ball_coordinates=Point3D(
                x=row["ball_x"], y=row["ball_y"], z=row.get("ball_z")
            ),
        )

        if transformer:
            frame = transformer.transform_frame(frame)

        return frame

    def deserialize(
        self, inputs: MetricaEPTSTrackingDataInputs
    ) -> TrackingDataset:
        with performance_logging("Loading metadata", logger=logger):
            metadata = load_metadata(inputs.meta_data)

            if metadata.provider and metadata.pitch_dimensions:
                transformer = self.get_transformer(
                    length=metadata.pitch_dimensions.length,
                    width=metadata.pitch_dimensions.width,
                    provider=metadata.coordinate_system.provider,
                )
            else:
                transformer = None

        with performance_logging("Loading data", logger=logger):
            # assume they are sorted
            frames = [
                self._frame_from_row(row, metadata, transformer)
                for row in read_raw_data(
                    raw_data=inputs.raw_data,
                    metadata=metadata,
                    sensor_ids=[
                        sensor.sensor_id for sensor in metadata.sensors
                    ],
                    sample_rate=self.sample_rate,
                    limit=self.limit,
                )
            ]

        if transformer:
            metadata = replace(
                metadata,
                pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
                coordinate_system=transformer.get_to_coordinate_system(),
            )

        return TrackingDataset(records=frames, metadata=metadata)



class KloppyError(Exception):
    pass


class DeserializationError(KloppyError):
    pass


class OrientationError(KloppyError):
    pass


class OrphanedRecordError(KloppyError):
    pass


class InvalidFilterError(KloppyError):
    pass


class AdapterError(KloppyError):
    pass


class InputNotFoundError(KloppyError):
    pass


class UnknownEncoderError(KloppyError):
    pass


class KloppyParameterError(KloppyError):
    pass



from dataclasses import dataclass, field
from typing import List, Dict, Optional, Callable, Union, Any

from kloppy.domain.models.common import DatasetType

from .common import Dataset, DataRecord, Player
from .pitch import Point


@dataclass
class PlayerData:
    coordinates: Point
    distance: Optional[float] = None
    speed: Optional[float] = None
    other_data: Dict[str, Any] = field(default_factory=dict)


@dataclass(repr=False)
class Frame(DataRecord):
    frame_id: int
    players_data: Dict[Player, PlayerData]
    other_data: Dict[str, Any]
    ball_coordinates: Point

    @property
    def record_id(self) -> int:
        return self.frame_id

    @property
    def players_coordinates(self):
        return {
            player: player_data.coordinates
            for player, player_data in self.players_data.items()
        }


@dataclass
class TrackingDataset(Dataset[Frame]):
    records: List[Frame]

    dataset_type: DatasetType = DatasetType.TRACKING

    @property
    def frames(self):
        return self.records

    @property
    def frame_rate(self):
        return self.metadata.frame_rate

    def to_pandas(
        self,
        record_converter: Callable[[Frame], Dict] = None,
        additional_columns: Dict[
            str, Union[Callable[[Frame], Any], Any]
        ] = None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:

            def record_converter(frame: Frame) -> Dict:
                row = dict(
                    period_id=frame.period.id if frame.period else None,
                    timestamp=frame.timestamp,
                    ball_state=frame.ball_state.value
                    if frame.ball_state
                    else None,
                    ball_owning_team_id=frame.ball_owning_team.team_id
                    if frame.ball_owning_team
                    else None,
                    ball_x=frame.ball_coordinates.x
                    if frame.ball_coordinates
                    else None,
                    ball_y=frame.ball_coordinates.y
                    if frame.ball_coordinates
                    else None,
                )
                for player, player_data in frame.players_data.items():
                    row.update(
                        {
                            f"{player.player_id}_x": player_data.coordinates.x,
                            f"{player.player_id}_y": player_data.coordinates.y,
                            f"{player.player_id}_d": player_data.distance,
                            f"{player.player_id}_s": player_data.speed,
                        }
                    )

                    if player_data.other_data:
                        for name, value in player_data.other_data.items():
                            row.update(
                                {
                                    f"{player.player_id}_{name}": value,
                                }
                            )

                if frame.other_data:
                    for name, value in frame.other_data.items():
                        row.update(
                            {
                                name: value,
                            }
                        )

                return row

        def generic_record_converter(frame: Frame):
            row = record_converter(frame)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(frame)
                    else:
                        value = v
                    row.update({k: value})

            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )


__all__ = ["Frame", "TrackingDataset", "PlayerData"]



import logging
import os

import pytest

from kloppy.domain import (
    AttackingDirection,
    Orientation,
    Provider,
    Point,
    Point3D,
    DatasetType,
)

from kloppy import secondspectrum


class TestSecondSpectrumTracking:
    @pytest.fixture
    def meta_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/second_spectrum_fake_metadata.xml"

    @pytest.fixture
    def raw_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/second_spectrum_fake_data.jsonl"

    @pytest.fixture
    def additional_meta_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/second_spectrum_fake_metadata.json"

    def test_correct_deserialization(
        self, meta_data: str, raw_data: str, additional_meta_data: str
    ):
        dataset = secondspectrum.load(
            meta_data=meta_data,
            raw_data=raw_data,
            additional_meta_data=additional_meta_data,
            only_alive=False,
            coordinates="secondspectrum",
        )

        # Check provider, type, shape, etc
        assert dataset.metadata.provider == Provider.SECONDSPECTRUM
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 376
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation == Orientation.FIXED_AWAY_HOME

        # Check the Periods
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == 0
        assert dataset.metadata.periods[0].end_timestamp == 2982240
        assert (
            dataset.metadata.periods[0].attacking_direction
            == AttackingDirection.AWAY_HOME
        )

        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == 3907360
        assert dataset.metadata.periods[1].end_timestamp == 6927840
        assert (
            dataset.metadata.periods[1].attacking_direction
            == AttackingDirection.HOME_AWAY
        )

        # Check some timestamps
        assert dataset.records[0].timestamp == 0  # First frame
        assert dataset.records[20].timestamp == 320.0  # Later frame

        # Check some players
        home_player = dataset.metadata.teams[0].players[2]
        assert home_player.player_id == "8xwx2"
        assert dataset.records[0].players_coordinates[home_player] == Point(
            x=-8.943903672572427, y=-28.171654132650364
        )

        away_player = dataset.metadata.teams[1].players[3]
        assert away_player.player_id == "2q0uv"
        assert dataset.records[0].players_coordinates[away_player] == Point(
            x=-45.11871334915762, y=-20.06459030559596
        )

        # Check the ball
        assert dataset.records[1].ball_coordinates == Point3D(
            x=-23.147073918432426, y=13.69367399756424, z=0.0
        )

        # Check pitch dimensions
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == -52.425
        assert pitch_dimensions.x_dim.max == 52.425
        assert pitch_dimensions.y_dim.min == -33.985
        assert pitch_dimensions.y_dim.max == 33.985

    def test_correct_normalized_deserialization(
        self, meta_data: str, raw_data: str, additional_meta_data: str
    ):
        dataset = secondspectrum.load(
            meta_data=meta_data,
            raw_data=raw_data,
            additional_meta_data=additional_meta_data,
            only_alive=False,
        )

        home_player = dataset.metadata.teams[0].players[2]
        assert dataset.records[0].players_coordinates[home_player] == Point(
            x=0.4146981051733674, y=0.9144718866065965
        )

        # Check normalised pitch dimensions
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == 0.0
        assert pitch_dimensions.x_dim.max == 1.0
        assert pitch_dimensions.y_dim.min == 0.0
        assert pitch_dimensions.y_dim.max == 1.0



import warnings
from typing import Union

from kloppy.infra.serializers.event.statsbomb import (
    StatsBombDeserializer,
    StatsbombInputs,
)
from kloppy.domain import EventDataset, Optional, List
from kloppy.io import open_as_file, FileLike


def load(
    event_data: FileLike,
    lineup_data: FileLike,
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
) -> EventDataset:
    """
    Load Statsbomb event data into a [`EventDataset`][kloppy.domain.models.event.EventDataset]

    Parameters:
        event_data: filename of json containing the events
        lineup_data: filename of json containing the lineup information
        event_types:
        coordinates:
    """
    deserializer = StatsBombDeserializer(
        event_types=event_types, coordinate_system=coordinates
    )
    with open_as_file(event_data) as event_data_fp, open_as_file(
        lineup_data
    ) as lineup_data_fp:

        return deserializer.deserialize(
            inputs=StatsbombInputs(
                event_data=event_data_fp, lineup_data=lineup_data_fp
            ),
        )


def load_open_data(
    match_id: Union[str, int] = "15946",
    event_types: Optional[List[str]] = None,
    coordinates: Optional[str] = None,
) -> EventDataset:
    warnings.warn(
        "\n\nYou are about to use StatsBomb public data."
        "\nBy using this data, you are agreeing to the user agreement. "
        "\nThe user agreement can be found here: https://github.com/statsbomb/open-data/blob/master/LICENSE.pdf"
        "\n"
    )

    return load(
        event_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/events/{match_id}.json",
        lineup_data=f"https://raw.githubusercontent.com/statsbomb/open-data/master/data/lineups/{match_id}.json",
        event_types=event_types,
        coordinates=coordinates,
    )



from dataclasses import replace, dataclass
from typing import Optional

from kloppy.domain import (
    FormationChangeEvent,
    Event,
    Ground,
    EventDataset,
    FormationType,
)
from ..builder import StateBuilder


@dataclass
class Formation:
    home: Optional[FormationType] = None
    away: Optional[FormationType] = None

    def __str__(self):
        if self.home and self.away:
            return f"{self.home.value} {self.away.value}"
        return "Unknown"


class FormationStateBuilder(StateBuilder):
    def initial_state(self, dataset: EventDataset) -> Formation:
        home_team, away_team = dataset.metadata.teams
        home_formation = home_team.starting_formation
        away_formation = away_team.starting_formation

        return Formation(home=home_formation, away=away_formation)

    def reduce_before(self, state: Formation, event: Event) -> Formation:
        return state

    def reduce_after(self, state: Formation, event: Event) -> Formation:
        if isinstance(event, FormationChangeEvent):
            if event.team.ground == Ground.HOME:
                state = replace(state, home=event.formation_type)
            else:
                state = replace(state, away=event.formation_type)
        return state



import logging
import os
import urllib.parse
from typing import Union, IO, BinaryIO, Tuple

from io import BytesIO

from kloppy.config import get_config
from kloppy.exceptions import InputNotFoundError
from kloppy.infra.io.adapters import get_adapter


logger = logging.getLogger(__name__)

_open = open


FileLike = Union[str, bytes, IO[bytes]]


def get_local_cache_stream(url: str, cache_dir: str) -> Tuple[BinaryIO, bool]:
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    filename = urllib.parse.quote_plus(url)
    local_filename = f"{cache_dir}/{filename}"

    # Open the file in append+read mode
    # this makes sure:
    # 1. The file is created when it does not exist
    # 2. The file is not truncated when it does exist
    # 3. The file can be read
    return _open(local_filename, "a+b"), (
        os.path.exists(local_filename)
        and os.path.getsize(local_filename) > 0
        and local_filename
    )


def open_as_file(input_: FileLike) -> IO:
    if isinstance(input_, str):
        if "{" in input_ or "<" in input_:
            return BytesIO(input_.encode("utf8"))
        else:
            adapter = get_adapter(input_)
            if adapter:
                cache_dir = get_config("cache")
                if cache_dir:
                    stream, local_cache_file = get_local_cache_stream(
                        input_, cache_dir
                    )
                else:
                    stream = BytesIO()
                    local_cache_file = None

                if not local_cache_file:
                    logger.info(f"Retrieving {input_}")
                    adapter.read_to_stream(input_, stream)
                    logger.info("Retrieval complete")
                else:
                    logger.info(f"Using local cached file {local_cache_file}")
                stream.seek(0)
            else:
                if not os.path.exists(input_):
                    raise InputNotFoundError(f"File {input_} does not exist")

                stream = _open(input_, "rb")
            return stream
    elif isinstance(input_, bytes):
        return BytesIO(input_)
    else:
        return input_



import shutil
from typing import BinaryIO

from kloppy.config import get_config
from kloppy.exceptions import AdapterError
from .adapter import Adapter


class S3Adapter(Adapter):
    def __init__(self):
        self.s3_fs = None

    def supports(self, url: str) -> bool:
        return url.startswith("s3://")

    def _get_s3fs(self):
        try:
            import s3fs
        except ImportError:
            raise AdapterError(
                "Seems like you don't have s3fs installed. Please"
                " install it using: pip install s3fs"
            )

        s3_fs = get_config("adapters.s3.s3fs")
        if not s3_fs:
            if not self.s3_fs:
                s3_fs = s3fs.S3FileSystem()
            else:
                s3_fs = self.s3_fs
        return s3_fs

    def read_to_stream(self, url: str, output: BinaryIO):
        s3_fs = self._get_s3fs()

        with s3_fs.open(url, "rb") as fp:
            shutil.copyfile(fp, output)



import sys

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import (
    Dict,
    List,
    Type,
    Union,
    Any,
    Callable,
    Optional,
    Iterable,
    overload,
    TYPE_CHECKING,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from kloppy.domain.models.common import DatasetType
from kloppy.utils import (
    camelcase_to_snakecase,
    removes_suffix,
    docstring_inherit_attributes,
)

from .common import DataRecord, Dataset, Player, Team
from .formation import FormationType
from .pitch import Point

from ...exceptions import OrphanedRecordError, InvalidFilterError

if TYPE_CHECKING:
    from ..services.transformers.event import Column


class ResultType(Enum):
    @property
    @abstractmethod
    def is_success(self):
        raise NotImplementedError

    def __str__(self):
        return self.value


class ShotResult(ResultType):
    """
    ShotResult

    Attributes:
        GOAL (ShotResult): Shot resulted in a goal
        OFF_TARGET (ShotResult): Shot was off target
        POST (ShotResult): Shot hit the post
        BLOCKED (ShotResult): Shot was blocked by another player
        SAVED (ShotResult): Shot was saved by the keeper
    """

    GOAL = "GOAL"
    OFF_TARGET = "OFF_TARGET"
    POST = "POST"
    BLOCKED = "BLOCKED"
    SAVED = "SAVED"
    OWN_GOAL = "OWN_GOAL"

    @property
    def is_success(self):
        """
        Returns if the shot was a goal
        """
        return self == self.GOAL


class PassResult(ResultType):
    """
    PassResult

    Attributes:
        COMPLETE (PassResult): Complete pass
        INCOMPLETE (PassResult): Incomplete pass (intercepted)
        OUT (PassResult): Ball went out
        OFFSIDE (PassResult): Offside
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"
    OFFSIDE = "OFFSIDE"

    @property
    def is_success(self):
        """
        Returns if the pass was complete
        """
        return self == self.COMPLETE


class TakeOnResult(ResultType):
    """
    TakeOnResult

    Attributes:
        COMPLETE (TakeOnResult): Complete take-on
        INCOMPLETE (TakeOnResult): Incomplete take-on
        OUT (TakeOnResult): Ball went out
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"

    @property
    def is_success(self):
        """
        Returns if the take-on was complete
        """
        return self == self.COMPLETE


class CarryResult(ResultType):
    """
    CarryResult

    Attributes:
        COMPLETE (CarryResult): Complete carry
        INCOMPLETE (CarryResult): Incomplete carry
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"

    @property
    def is_success(self):
        """
        Returns if the carry was complete
        """
        return self == self.COMPLETE


class CardType(Enum):
    """
    CardType

    Attributes:
        FIRST_YELLOW (CardType): First yellow card
        SECOND_YELLOW (CardType): Second yellow card
        RED (CardType): Red card
    """

    FIRST_YELLOW = "FIRST_YELLOW"
    SECOND_YELLOW = "SECOND_YELLOW"
    RED = "RED"


class EventType(Enum):
    """
    Attributes:
        GENERIC (EventType): Unrecognised event type
        PASS (EventType):
        SHOT (EventType):
        TAKE_ON (EventType):
        CARRY (EventType):
        SUBSTITUTION (EventType):
        CARD (EventType):
        PLAYER_ON (EventType):
        PLAYER_OFF (EventType):
        RECOVERY (EventType):
        BALL_OUT (EventType):
        FOUL_COMMITTED (EventType):
        FORMATION_CHANGE (EventType):
    """

    GENERIC = "generic"

    PASS = "PASS"
    SHOT = "SHOT"
    TAKE_ON = "TAKE_ON"
    CARRY = "CARRY"
    SUBSTITUTION = "SUBSTITUTION"
    CARD = "CARD"
    PLAYER_ON = "PLAYER_ON"
    PLAYER_OFF = "PLAYER_OFF"
    RECOVERY = "RECOVERY"
    BALL_OUT = "BALL_OUT"
    FOUL_COMMITTED = "FOUL_COMMITTED"
    FORMATION_CHANGE = "FORMATION_CHANGE"

    def __repr__(self):
        return self.value


@dataclass
class Qualifier(ABC):
    value: None

    @abstractmethod
    def to_dict(self):
        """
        Return the qualifier as a dict
        """
        pass

    @property
    def name(self):
        return camelcase_to_snakecase(
            removes_suffix(type(self).__name__, "Qualifier")
        )


@dataclass
class BoolQualifier(Qualifier, ABC):
    value: bool

    def to_dict(self):
        return {f"is_{self.name}": self.value}


class EnumQualifier(Qualifier, ABC):
    value: Enum

    def to_dict(self):
        return {f"{self.name}_type": self.value.value}


class SetPieceType(Enum):
    """
    SetPieceType

    Attributes:
        GOAL_KICK (SetPieceType):
        FREE_KICK (SetPieceType):
        THROW_IN (SetPieceType):
        CORNER_KICK (SetPieceType):
        PENALTY (SetPieceType):
        KICK_OFF (SetPieceType):
    """

    GOAL_KICK = "GOAL_KICK"
    FREE_KICK = "FREE_KICK"
    THROW_IN = "THROW_IN"
    CORNER_KICK = "CORNER_KICK"
    PENALTY = "PENALTY"
    KICK_OFF = "KICK_OFF"


@dataclass
class SetPieceQualifier(EnumQualifier):
    """
    SetPieceQualifier

    Attributes:
        value: Specifies the type of set piece
    """

    value: SetPieceType


@dataclass
class CardQualifier(EnumQualifier):
    """
    CardQualifier

    Attributes:
        value: Specifies the type card
    """

    value: CardType


class PassType(Enum):
    """
    PassType

    Attributes:
        CROSS (PassType):
        HAND_PASS (PassType):
        HEAD_PASS (PassType):
        HIGH_PASS (PassType):
        LAUNCH (PassType):
        SIMPLE_PASS (PassType):
        SMART_PASS (PassType):
        LONG_BALL (PassType)
        THROUGH_BALL (PassType)
        CHIPPED_PASS (PassType)
        FLICK_ON (PassType)
        ASSIST (PassType)
        ASSIST_2ND (PassType)
        SWITCH_OF_PLAY (PassType)
    """

    CROSS = "CROSS"
    HAND_PASS = "HAND_PASS"
    HEAD_PASS = "HEAD_PASS"
    HIGH_PASS = "HIGH_PASS"
    LAUNCH = "LAUNCH"
    SIMPLE_PASS = "SIMPLE_PASS"
    SMART_PASS = "SMART_PASS"
    LONG_BALL = "LONG_BALL"
    THROUGH_BALL = "THROUGH_BALL"
    CHIPPED_PASS = "CHIPPED_PASS"
    FLICK_ON = "FLICK_ON"
    ASSIST = "ASSIST"
    ASSIST_2ND = "ASSIST_2ND"
    SWITCH_OF_PLAY = "SWITCH_OF_PLAY"


@dataclass
class PassQualifier(EnumQualifier):
    value: PassType


class BodyPart(Enum):
    """
    BodyPart

    Attributes:
        RIGHT_FOOT (BodyPart): Pass or Shot with right foot, save with right foot (for goalkeepers).
        LEFT_FOOT (BodyPart): Pass or Shot with leftt foot, save with left foot (for goalkeepers).
        HEAD (BodyPart): Pass or Shot with head, save with head (for goalkeepers).
        BOTH_HANDS (BodyPart): Goalkeeper only. Save with both hands.
        CHEST (BodyPart): Goalkeeper only. Save with chest.
        LEFT_HAND (BodyPart): Goalkeeper only. Save with left hand.
        RIGHT_HAND (BodyPart): Goalkeeper only. Save with right hand.
        DROP_KICK (BodyPart): Pass is a keeper drop kick.
        KEEPER_ARM (BodyPart): Pass thrown from keepers hands.
        OTHER (BodyPart): Other body part (chest, back, etc.), for Pass and Shot.
        NO_TOUCH (BodyPart): Pass only. A player deliberately let the pass go past him
                             instead of receiving it to deliver to a teammate behind him.
                             (Also known as a "dummy").
    """

    RIGHT_FOOT = "RIGHT_FOOT"
    LEFT_FOOT = "LEFT_FOOT"
    HEAD = "HEAD"

    BOTH_HANDS = "BOTH_HANDS"
    CHEST = "CHEST"
    LEFT_HAND = "LEFT_HAND"
    RIGHT_HAND = "RIGHT_HAND"
    DROP_KICK = "DROP_KICK"
    KEEPER_ARM = "KEEPER_ARM"
    OTHER = "OTHER"
    NO_TOUCH = "NO_TOUCH"


@dataclass
class BodyPartQualifier(EnumQualifier):
    value: BodyPart


class GoalkeeperAction(Enum):
    REFLEX = "REFLEX"
    SAVE_ATTEMPT = "SAVE_ATTEMPT"


@dataclass
class GoalkeeperActionQualifier(EnumQualifier):
    value: GoalkeeperAction


@dataclass
class CounterAttackQualifier(BoolQualifier):
    pass


@dataclass
@docstring_inherit_attributes(DataRecord)
class Event(DataRecord, ABC):
    """
    Abstract event baseclass. All other event classes inherit from this class.

    Attributes:
        event_id: identifier given by provider
        team: See [`Team`][kloppy.domain.models.common.Team]
        player: See [`Player`][kloppy.domain.models.common.Player]
        coordinates: Coordinates where event happened. See [`Point`][kloppy.domain.models.pitch.Point]
        raw_event: Dict
        state: Dict[str, Any]
        qualifiers: See [`Qualifier`][kloppy.domain.models.event.Qualifier]
    """

    event_id: str
    team: Team
    player: Player
    coordinates: Point

    result: Optional[ResultType]

    raw_event: Dict
    state: Dict[str, Any]
    related_event_ids: List[str]

    qualifiers: List[Qualifier]

    @property
    def record_id(self) -> str:
        return self.event_id

    @property
    @abstractmethod
    def event_type(self) -> EventType:
        raise NotImplementedError

    @property
    @abstractmethod
    def event_name(self) -> str:
        raise NotImplementedError

    @classmethod
    def create(cls, **kwargs):
        extra_kwargs = {"state": {}}
        if "related_event_ids" not in kwargs:
            extra_kwargs["related_event_ids"] = []
        return cls(**kwargs, **extra_kwargs)

    def get_qualifier_value(self, qualifier_type: Type[Qualifier]):
        """
        Returns the Qualifier of a certain type, or None if qualifier is not present.

        Arguments:
            qualifier_type: one of the following QualifierTypes: [`SetPieceQualifier`][kloppy.domain.models.event.SetPieceQualifier]
                [`BodyPartQualifier`][kloppy.domain.models.event.BodyPartQualifier] [`PassQualifier`][kloppy.domain.models.event.PassQualifier]

        Examples:
            >>> from kloppy.domain import SetPieceQualifier
            >>> pass_event.get_qualifier_value(SetPieceQualifier)
            <SetPieceType.GOAL_KICK: 'GOAL_KICK'>
        """
        if self.qualifiers:
            for qualifier in self.qualifiers:
                if isinstance(qualifier, qualifier_type):
                    return qualifier.value
        return None

    def get_related_events(self) -> List["Event"]:
        if not self.dataset:
            raise OrphanedRecordError()

        return [
            self.dataset.get_record_by_id(event_id)
            for event_id in self.related_event_ids
        ]

    def get_related_event(
        self, type_: Union[str, EventType]
    ) -> Optional["Event"]:
        event_type = (
            EventType[type_.upper()] if isinstance(type_, str) else type_
        )
        for related_event in self.get_related_events():
            if related_event.event_type == event_type:
                return related_event
        return None

    """Define all related events for easy access"""

    def related_pass(self) -> Optional["PassEvent"]:
        return self.get_related_event(EventType.PASS)

    def related_shot(self) -> Optional["ShotEvent"]:
        return self.get_related_event(EventType.SHOT)

    def related_take_on(self) -> Optional["TakeOnEvent"]:
        return self.get_related_event(EventType.TAKE_ON)

    def related_carry(self) -> Optional["CarryEvent"]:
        return self.get_related_event(EventType.CARRY)

    def related_substitution(self) -> Optional["SubstitutionEvent"]:
        return self.get_related_event(EventType.SUBSTITUTION)

    def related_card(self) -> Optional["CardEvent"]:
        return self.get_related_event(EventType.CARD)

    def related_player_on(self) -> Optional["PlayerOnEvent"]:
        return self.get_related_event(EventType.PLAYER_ON)

    def related_player_off(self) -> Optional["PlayerOffEvent"]:
        return self.get_related_event(EventType.PLAYER_OFF)

    def related_recovery(self) -> Optional["RecoveryEvent"]:
        return self.get_related_event(EventType.RECOVERY)

    def related_ball_out(self) -> Optional["BallOutEvent"]:
        return self.get_related_event(EventType.BALL_OUT)

    def related_foul_committed(self) -> Optional["FoulCommittedEvent"]:
        return self.get_related_event(EventType.FOUL_COMMITTED)

    def related_formation_change(self) -> Optional["FormationChangeEvent"]:
        return self.get_related_event(EventType.FORMATION_CHANGE)

    def matches(self, filter_) -> bool:
        if filter_ is None:
            return True
        elif callable(filter_):
            return filter_(self)
        elif isinstance(filter_, str):
            """
            Allowed formats:
            1. <event_type>
            2. <event_type>.<result>

            This format always us to go to css selectors without breaking existing code.
            """
            parts = filter_.upper().split(".")
            if len(parts) == 2:
                event_type, result = parts
            elif len(parts) == 1:
                event_type = parts[0]
                result = None
            else:
                raise InvalidFilterError(
                    f"Don't know how to apply filter {filter_}"
                )

            if event_type:
                try:
                    if self.event_type != EventType[event_type]:
                        return False
                except KeyError:
                    raise InvalidFilterError(
                        f"Cannot find event type {event_type}. Possible options: {[e.value.lower() for e in EventType]}"
                    )

            if result:
                if not self.result:
                    return False

                try:
                    if self.result != self.result.__class__[result]:
                        return False
                except KeyError:
                    # result isn't applicable for this event
                    # example: result='GOAL' event=<Pass>
                    return False

            return True

    def __str__(self):
        m, s = divmod(self.timestamp, 60)

        event_type = (
            self.__class__.__name__
            if not isinstance(self, GenericEvent)
            else f"GenericEvent:{self.event_name}"
        )

        return (
            f"<{event_type} "
            f"event_id='{self.event_id}' "
            f"time='P{self.period.id}T{m:02.0f}:{s:02.0f}' "
            f"player='{self.player}' "
            f"result='{self.result}'>"
        )

    def __repr__(self):
        return str(self)


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class GenericEvent(Event):
    """
    GenericEvent

    Attributes:
        event_type (EventType): `EventType.GENERIC` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"generic"`
    """

    event_type: EventType = EventType.GENERIC
    event_name: str = "generic"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class ShotEvent(Event):
    """
    ShotEvent

    Attributes:
        event_type (EventType): `EventType.SHOT` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"shot"`,
        result_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        result (ShotResult): See [`ShotResult`][kloppy.domain.models.event.ShotResult]
    """

    result: ShotResult
    result_coordinates: Point = None

    event_type: EventType = EventType.SHOT
    event_name: str = "shot"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PassEvent(Event):
    """
    PassEvent

    Attributes:
        event_type (EventType): `EventType.PASS` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"pass"`
        receive_timestamp (float):
        receiver_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        receiver_player (Player): See [`Player`][kloppy.domain.models.common.Player]
        result (PassResult): See [`PassResult`][kloppy.domain.models.event.PassResult]
    """

    receive_timestamp: float
    receiver_player: Player
    receiver_coordinates: Point

    result: PassResult

    event_type: EventType = EventType.PASS
    event_name: str = "pass"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class TakeOnEvent(Event):
    """
    TakeOnEvent

    Attributes:
        event_type (EventType): `EventType.TAKE_ON` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"take_on"`,
        result (TakeOnResult): See [`TakeOnResult`][kloppy.domain.models.event.TakeOnResult]
    """

    result: TakeOnResult

    event_type: EventType = EventType.TAKE_ON
    event_name: str = "take_on"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class CarryEvent(Event):
    """
    CarryEvent

    Attributes:
        event_type (EventType): `EventType.CARRY` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"carry"`,
        end_timestamp (float):
        end_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        result (CarryResult): See [`CarryResult`][kloppy.domain.models.event.CarryResult]
    """

    end_timestamp: float
    end_coordinates: Point

    result: CarryResult

    event_type: EventType = EventType.CARRY
    event_name: str = "carry"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class SubstitutionEvent(Event):
    """
    SubstitutionEvent

    Attributes:
        event_type (EventType): `EventType.SUBSTITUTION` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"substitution"`,
        replacement_player (Player): See [`Player`][kloppy.domain.models.common.Player]
    """

    replacement_player: Player

    event_type: EventType = EventType.SUBSTITUTION
    event_name: str = "substitution"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PlayerOffEvent(Event):
    """
    PlayerOffEvent

    Attributes:
        event_type (EventType): `EventType.PLAYER_OFF` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"player_off"`
    """

    event_type: EventType = EventType.PLAYER_OFF
    event_name: str = "player_off"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PlayerOnEvent(Event):
    """
    PlayerOnEvent

    Attributes:
        event_type (EventType): `EventType.PLAYER_ON` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"player_on"`
    """

    event_type: EventType = EventType.PLAYER_ON
    event_name: str = "player_on"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class CardEvent(Event):
    """
    CardEvent

    Attributes:
        event_type (EventType): `EventType.CARD` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"card"`
        card_type: See [`CardType`][kloppy.domain.models.event.CardType]
    """

    card_type: CardType

    event_type: EventType = EventType.CARD
    event_name: str = "card"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class FormationChangeEvent(Event):
    """
    FormationChangeEvent

    Attributes:
        event_type (EventType): `EventType.FORMATION_CHANGE` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"card"`
        formation_type: See [`FormationType`][kloppy.domain.models.formation.FormationType]
    """

    formation_type: FormationType

    event_type: EventType = EventType.FORMATION_CHANGE
    event_name: str = "formation_change"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class RecoveryEvent(Event):
    """
    RecoveryEvent

    Attributes:
        event_type (EventType): `EventType.RECOVERY` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "recovery"
    """

    event_type: EventType = EventType.RECOVERY
    event_name: str = "recovery"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class BallOutEvent(Event):
    """
    BallOutEvent

    Attributes:
        event_type (EventType): `EventType.BALL_OUT` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "ball_out"
    """

    event_type: EventType = EventType.BALL_OUT
    event_name: str = "ball_out"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class FoulCommittedEvent(Event):
    """
    FoulCommittedEvent

    Attributes:
        event_type (EventType): `EventType.FOUL_COMMITTED` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "foul_committed"
    """

    event_type: EventType = EventType.FOUL_COMMITTED
    event_name: str = "foul_committed"


@dataclass(repr=False)
class EventDataset(Dataset[Event]):
    """
    EventDataset

    Attributes:
        metadata: See [`Metadata`][kloppy.domain.models.common.Metadata]
        records (List[Event]): See [`Event`][kloppy.domain.models.event.Event]
        dataset_type: `DatasetType.EVENT` (See [`DatasetType`][kloppy.domain.models.common.DatasetType])
        events: alias for `records`
    """

    records: List[Event]

    dataset_type: DatasetType = DatasetType.EVENT

    @property
    def events(self):
        return self.records

    def get_event_by_id(self, event_id: str) -> Event:
        return self.get_record_by_id(event_id)

    def add_state(self, *builder_keys):
        """
        See [add_state][kloppy.domain.services.state_builder.add_state]
        """
        from kloppy.domain.services.state_builder import add_state

        return add_state(self, *builder_keys)

    def to_pandas(
        self,
        record_converter: Callable[[Event], Dict] = None,
        additional_columns: Dict[
            str, Union[Callable[[Event], Any], Any]
        ] = None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:
            from ..services.transformers.attribute import DefaultTransformer

            record_converter = DefaultTransformer()

        def generic_record_converter(event: Event):
            row = record_converter(event)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(event)
                    else:
                        value = v
                    row.update({k: value})
            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )

    @overload
    def to_records(
        self,
        *columns: "Column",
        as_list: Literal[True] = True,
        **named_columns: "Column",
    ) -> List[Dict[str, Any]]:
        ...

    @overload
    def to_records(
        self,
        *columns: "Column",
        as_list: Literal[False] = False,
        **named_columns: "Column",
    ) -> Iterable[Dict[str, Any]]:
        ...

    def to_records(
        self,
        *columns: "Column",
        as_list: bool = True,
        **named_columns: "Column",
    ) -> Union[List[Dict[str, Any]], Iterable[Dict[str, Any]]]:
        from ..services.transformers.event import EventToRecordTransformer

        transformer = EventToRecordTransformer(*columns, **named_columns)
        iterator = map(transformer, self.events)
        if as_list:
            return list(iterator)
        else:
            return iterator

    def to_df(self, *columns: "Column", **named_columns: "Column"):
        from pandas import DataFrame

        return DataFrame.from_records(
            self.to_records(*columns, **named_columns, as_list=False)
        )


__all__ = [
    "EnumQualifier",
    "ResultType",
    "EventType",
    "ShotResult",
    "PassResult",
    "TakeOnResult",
    "CarryResult",
    "Event",
    "GenericEvent",
    "ShotEvent",
    "PassEvent",
    "TakeOnEvent",
    "CarryEvent",
    "SubstitutionEvent",
    "PlayerOnEvent",
    "PlayerOffEvent",
    "CardEvent",
    "CardType",
    "CardQualifier",
    "FormationType",
    "FormationChangeEvent",
    "EventDataset",
    "RecoveryEvent",
    "FoulCommittedEvent",
    "BallOutEvent",
    "SetPieceType",
    "Qualifier",
    "SetPieceQualifier",
    "PassQualifier",
    "PassType",
    "BodyPart",
    "BodyPartQualifier",
    "GoalkeeperAction",
    "GoalkeeperActionQualifier",
    "CounterAttackQualifier",
]



import os
from itertools import groupby

from kloppy.domain import EventType, Event, EventDataset, FormationType
from kloppy.domain.services.state_builder.builder import StateBuilder, T
from kloppy.utils import performance_logging
from kloppy import statsbomb


class TestStateBuilder:
    """"""

    def _load_dataset(self, base_filename="statsbomb"):
        base_dir = os.path.dirname(__file__)

        return statsbomb.load(
            event_data=f"{base_dir}/files/{base_filename}_event.json",
            lineup_data=f"{base_dir}/files/{base_filename}_lineup.json",
        )

    def test_score_state_builder(self):
        dataset = self._load_dataset()

        with performance_logging("add_state"):
            dataset_with_state = dataset.add_state("score")

        events_per_score = {}
        for score, events in groupby(
            dataset_with_state.events, lambda event: event.state["score"]
        ):
            events = list(events)
            events_per_score[str(score)] = len(events)

        assert events_per_score == {
            "0-0": 2898,
            "1-0": 717,
            "2-0": 405,
            "3-0": 3,
        }

    def test_sequence_state_builder(self):
        dataset = self._load_dataset()

        with performance_logging("add_state"):
            dataset_with_state = dataset.add_state("sequence")

        events_per_sequence = {}
        for sequence_id, events in groupby(
            dataset_with_state.events,
            lambda event: event.state["sequence"].sequence_id,
        ):
            events = list(events)
            events_per_sequence[sequence_id] = len(events)

        assert events_per_sequence[0] == 4
        assert events_per_sequence[51] == 14

    def test_lineup_state_builder(self):
        dataset = self._load_dataset("statsbomb_15986")

        with performance_logging("add_state"):
            dataset_with_state = dataset.add_state("lineup")

        last_events = []
        for lineup, events in groupby(
            dataset_with_state.events, lambda event: event.state["lineup"]
        ):
            events = list(events)
            # inspect last event which changed the lineup
            last_events.append((events[-1].event_type, len(lineup.players)))

        assert last_events == [
            (EventType.CARD, 22),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.GENERIC, 21),
        ]

    def test_formation_state_builder(self):
        dataset = self._load_dataset("statsbomb")

        with performance_logging("add_state"):
            dataset_with_state = dataset.add_state("formation")

        events_per_formation_change = {}
        for formation, events in groupby(
            dataset_with_state.events,
            lambda event: event.state["formation"].away,
        ):
            events = list(events)
            events_per_formation_change[str(formation)] = len(events)

        # inspect FormationChangeEvent usage and formation state_builder
        assert events_per_formation_change["4-1-4-1"] == 3074
        assert events_per_formation_change["4-4-2"] == 949

        assert dataset.metadata.teams[0].starting_formation == FormationType(
            "4-4-2"
        )
        assert dataset_with_state.events[1].state[
            "formation"
        ].home == FormationType("4-4-2")

    def test_register_custom_builder(self):
        class CustomStateBuilder(StateBuilder):
            def initial_state(self, dataset: EventDataset) -> int:
                return 0

            def reduce_before(self, state: int, event: Event) -> int:
                return state + 1

            def reduce_after(self, state: int, event: Event) -> int:
                return state + 1

        dataset = self._load_dataset("statsbomb_15986")

        with performance_logging("add_state"):
            dataset_with_state = dataset.add_state("custom")

        assert dataset_with_state.events[0].state["custom"] == 1
        assert dataset_with_state.events[1].state["custom"] == 3
        assert dataset_with_state.events[2].state["custom"] == 5
        assert dataset_with_state.events[3].state["custom"] == 7



import os

import pytest
from kloppy import opta
from kloppy.config import set_config, get_config, config_context, reset_config
from kloppy.domain import KloppyCoordinateSystem, OptaCoordinateSystem


class TestConfig:
    @pytest.fixture(autouse=True)
    def tear_down(self):
        """Make sure all config changes are reverted for next tests."""
        try:
            yield
        finally:
            reset_config()

    @pytest.fixture
    def f24_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/opta_f24.xml"

    @pytest.fixture
    def f7_data(self) -> str:
        base_dir = os.path.dirname(__file__)
        return f"{base_dir}/files/opta_f7.xml"

    def test_set_config(self, f24_data: str, f7_data: str):
        """
        Make sure the coordinate system can be set and is used
        when loading data.
        """
        dataset = opta.load(f24_data=f24_data, f7_data=f7_data)
        assert isinstance(
            dataset.metadata.coordinate_system, KloppyCoordinateSystem
        )

        set_config("coordinate_system", "opta")

        dataset = opta.load(f24_data=f24_data, f7_data=f7_data)
        assert isinstance(
            dataset.metadata.coordinate_system, OptaCoordinateSystem
        )

    def test_config_context(self, f24_data: str, f7_data: str):
        assert get_config("coordinate_system") == "kloppy"

        with config_context("coordinate_system", "opta"):
            assert get_config("coordinate_system") == "opta"

            dataset = opta.load(f24_data=f24_data, f7_data=f7_data)

        assert get_config("coordinate_system") == "kloppy"

        assert isinstance(
            dataset.metadata.coordinate_system, OptaCoordinateSystem
        )



from abc import abstractmethod, ABC
from typing import BinaryIO


class Adapter(ABC):
    @abstractmethod
    def supports(self, url: str) -> bool:
        pass

    @abstractmethod
    def read_to_stream(self, url: str, output: BinaryIO):
        pass


__all__ = ["Adapter"]



from fnmatch import fnmatch
from typing import NewType, Union, Callable, Any, Dict

from kloppy.domain import Event
from kloppy.domain.services.transformers.attribute import DefaultTransformer
from kloppy.exceptions import KloppyError

Column = NewType("Column", Union[str, Callable[[Event], Any]])


class EventToRecordTransformer:
    def __init__(
        self,
        *columns: Column,
        **named_columns: Column,
    ):
        if not columns and not named_columns:

            converter = DefaultTransformer()
        else:
            default = DefaultTransformer()
            has_string_columns = any(
                not callable(column) for column in columns
            )

            def converter(event: Event) -> Dict[str, Any]:
                if has_string_columns:
                    default_row = default(event)
                else:
                    default_row = {}

                row = {}
                for column in columns:
                    if callable(column):
                        res = column(event)
                        if not isinstance(res, dict):
                            raise KloppyError(
                                f"A function column should return a dictionary"
                            )
                        row.update(res)
                    else:
                        if column == "*":
                            row.update(default_row)
                        elif "*" in column:
                            row.update(
                                {
                                    k: v
                                    for k, v in default_row.items()
                                    if fnmatch(k, column)
                                }
                            )
                        elif column in default_row:
                            row[column] = default_row[column]
                        else:
                            row[column] = getattr(event, column, None)

                for name, column in named_columns.items():
                    row[name] = column(event) if callable(column) else column

                return row

        self.converter = converter

    def __call__(self, event: Event) -> Dict[str, Any]:
        return self.converter(event)
